@@ -1,5 +1,5 @@
 from typing import Dict
-
+import random
 from negmas import MechanismState, SAOState, SAOResponse, ResponseType
 
 from .agent import OneShotAgent, OneShotSyncAgent
@@ -9,6 +9,9 @@ __all__ = [
     "SyncRandomOneShotAgent",
 ]
 
+PROB_ACCEPTANCE = 0.1
+PROB_END = 0.005
+
 
 class RandomOneShotAgent(OneShotAgent):
     def _random_offer(self, negotiator_id: str):
@@ -17,6 +20,13 @@ class RandomOneShotAgent(OneShotAgent):
     def propose(self, negotiator_id: str, state: MechanismState) -> "Outcome":
         return self._random_offer(negotiator_id)
 
+    def respond(self, negotiator_id, state, offer):
+        if random.random() < PROB_END:
+            return ResponseType.END_NEGOTIATION
+        if random.random() < PROB_ACCEPTANCE:
+            return ResponseType.ACCEPT_OFFER
+        return ResponseType.REJECT_OFFER
+
 
 class SyncRandomOneShotAgent(OneShotSyncAgent):
     def _random_offer(self, negotiator_id: str):
@@ -27,8 +37,11 @@ class SyncRandomOneShotAgent(OneShotSyncAgent):
     ) -> Dict[str, SAOResponse]:
         proposals = dict()
         for id in self.negotiators.keys():
-            proposals[id] = SAOResponse(
-                ResponseType.REJECT_OFFER, self._random_offer(id)
+
+            proposals[id] = (
+                SAOResponse(ResponseType.ACCEPT_OFFER, None)
+                if random.random() < PROB_ACCEPTANCE
+                else SAOResponse(ResponseType.REJECT_OFFER, self._random_offer(id))
             )
         return proposals
 
