@@ -2,6 +2,7 @@
 Implements the one shot version of SCML
 """
 import sys
+import math
 import random
 import logging
 import itertools
@@ -20,6 +21,7 @@ from typing import (
     Tuple,
     Union,
     Callable,
+    Set,
 )
 import numpy as np
 from matplotlib.axis import Axis
@@ -27,6 +29,7 @@ import networkx as nx
 from negmas import (
     World,
     Agent,
+    Adapter,
     AgentWorldInterface,
     Issue,
     AgentMechanismInterface,
@@ -37,7 +40,8 @@ from negmas import (
     Breach,
     PassThroughSAONegotiator,
     SAOController,
-    NoContractExecutionMixin,
+    # NoContractExecutionMixin,
+    TimeInAgreementMixin,
     Operations,
     BreachProcessing,
     DEFAULT_EDGE_TYPES,
@@ -346,9 +350,15 @@ class OneShotAWI(AgentWorldInterface):
         return self._world.agent_delivery_penalty[self.agent.id]
 
 
-class _OneShotBaseAgent(Agent):
+class _OneShotAdapter(Adapter):
     """The base class of all one-shot agents"""
 
+    def on_negotiation_failure(self, partners, annotation, mechanism, state):
+        return self._obj.on_negotiation_failure(partners, annotation, mechanism, state)
+
+    def on_negotiation_success(self, contract, mechanism):
+        return self._obj.on_negotiation_success(contract, mechanism)
+
     def on_contract_executed(self, contract: Contract) -> None:
         pass
 
@@ -357,18 +367,6 @@ class _OneShotBaseAgent(Agent):
     ) -> None:
         pass
 
-    def __init__(
-        self,
-        controller_type: Union[str, Type["OneShotAgent"]],
-        controller_params: Optional[Dict[str, Any]] = None,
-        **kwargs,
-    ):
-        super().__init__(**kwargs)
-        self.controller_type = controller_type
-        self.controller_params = dict() if not controller_params else controller_params
-        self.controller = None
-        self.profile = None
-
     def make_ufun(self):
         awi: OneShotAWI = self.awi
         qin, pin = (
@@ -390,21 +388,9 @@ class _OneShotBaseAgent(Agent):
             delivery_penalty=awi.current_delivery_penalty,
         )
 
-    def make_controller(self) -> Optional[SAOController]:
-        self.controller = instantiate(
-            self.controller_type,
-            owner=self,
-            ufun=self.make_ufun(),
-            **self.controller_params,
-        )
-        return self.controller
-
     def init(self):
-        self.controller = self.make_controller()
-        self.controller.init()
-
-    def step(self):
-        self.controller.step()
+        self.ufun = self.make_ufun()
+        super().init()
 
     def to_dict(self):
         return {
@@ -426,9 +412,7 @@ class _OneShotBaseAgent(Agent):
         req_id: Optional[str],
     ) -> Optional[Negotiator]:
         partner = [_ for _ in partners if _ != self.id][0]
-        if not self.controller:
-            self.controller = self.make_controller()
-        neg = self.controller.create_negotiator(PassThroughSAONegotiator, name=partner)
+        neg = self._obj.create_negotiator(PassThroughSAONegotiator, name=partner)
         return neg
 
     def set_renegotiation_agenda(
@@ -447,39 +431,23 @@ class _OneShotBaseAgent(Agent):
     def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
         pass
 
-    @property
-    def internal_state(self) -> Dict[str, Any]:
-        """Returns the internal state of the agent for debugging purposes"""
-        return self.controller.internal_state if self.controller else dict()
 
-    def on_negotiation_failure(self, *args, **kwargs) -> None:
-        """Called whenever a negotiation ends without agreement"""
-        return (
-            self.controller.on_negotiation_failure(*args, **kwargs)
-            if self.controller
-            else None
-        )
-
-    def on_negotiation_success(self, *args, **kwargs) -> None:
-        """Called whenever a negotiation ends with agreement"""
-        return (
-            self.controller.on_negotiation_success(*args, **kwargs)
-            if self.controller
-            else None
-        )
-
-    def sign_all_contracts(self, contracts: List[Contract]) -> List[Optional[str]]:
-        """Signs all contracts"""
-        return [self.id] * len(contracts)
-
-
-class _SystemAgent(_OneShotBaseAgent):
+class _SystemAgent(_OneShotAdapter):
     """Implements an agent for handling system operations"""
 
     def __init__(self, *args, role, **kwargs):
         super().__init__(*args, **kwargs)
         self.id = role
         self.name = role
+        self.profile = None
+
+    @property
+    def type_name(self):
+        return "System"
+
+    @property
+    def short_type_name(self):
+        return "System"
 
     def respond_to_negotiation_request(
         self,
@@ -515,7 +483,7 @@ class _SystemAgent(_OneShotBaseAgent):
         return [self.id] * len(contracts)
 
 
-class SCML2020OneShotWorld(NoContractExecutionMixin, World):
+class SCML2020OneShotWorld(TimeInAgreementMixin, World):
     """Implements the SCML-OneShot variant of the SCM world.
 
     Args:
@@ -589,7 +557,9 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         # debugging parameters
         agent_name_reveals_position: bool = True,
         agent_name_reveals_type: bool = True,
-        # evaluation paramters
+        # evaluation paramters (for compatibility with SCML2020World)
+        inventory_valuation_catalog=0,
+        inventory_valuation_trading=0,
         **kwargs,
     ):
         self._profits: Dict[str, List[float]] = defaultdict(list)
@@ -676,12 +646,15 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         process_inputs = [[i] for i in range(n_processes)]
         process_outputs = [[i + 1] for i in range(n_processes)]
         self.exogenous_dynamic = exogenous_dynamic
+        agent_params = (
+            [copy.deepcopy(_) for _ in agent_params] if agent_params else agent_params
+        )
         self.info.update(
             process_inputs=process_inputs,
             process_outputs=process_outputs,
             catalog_prices=catalog_prices,
             agent_types_final=[get_full_type_name(_) for _ in agent_types],
-            agent_params_final=agent_params,
+            agent_params_final=[copy.deepcopy(_) for _ in agent_params],
             initial_balance_final=initial_balance,
             bankruptcy_limit=bankruptcy_limit,
             financial_report_period=financial_report_period,
@@ -704,6 +677,7 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             publish_trading_prices=publish_trading_prices,
             selected_price_multiplier=price_multiplier,
         )
+        TimeInAgreementMixin.init(self, time_field="time")
         self.bulletin_board.add_section("reports_agent")
         self.bulletin_board.add_section("reports_time")
         if self.publish_exogenous_summary:
@@ -736,10 +710,16 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         )
         # breakpoint()
         agent_types = [get_class(_) for _ in agent_types]
+        for p in agent_params:
+            p["obj"] = get_class(p["controller_type"])(
+                **p.get("controller_params", dict())
+            )
+            del p["controller_type"]
+            if "controller_params" in p.keys():
+                del p["controller_params"]
+
         self.controller_types = [
-            get_class(_["controller_type"])._type_name()
-            if _["controller_type"]
-            else "system_agent"
+            get_class(_["obj"])._type_name() if _["obj"] else "system_agent"
             for _ in agent_params
         ]
         assert (
@@ -752,8 +732,8 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         else:
             default_names = [unique_name("", add_time=False) for _ in range(n_agents)]
         if agent_name_reveals_type:
-            for i, at in enumerate(self.controller_types):
-                default_names[i] += f"{get_class(at).__name__[:3]}"
+            for i, at in enumerate(agent_params):
+                default_names[i] += f"{get_class(at['obj']).__class__.__name__[:3]}"
         agent_levels = [p.level for p in profiles]
         if agent_name_reveals_position:
             for i, l in enumerate(agent_levels):
@@ -781,8 +761,8 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
                     agent_params[i]["name"] = ns
         agent_types += [_SystemAgent, _SystemAgent]
         agent_params += [
-            {"role": SYSTEM_SELLER_ID, "controller_type": None},
-            {"role": SYSTEM_BUYER_ID, "controller_type": None},
+            {"role": SYSTEM_SELLER_ID, "obj": None},
+            {"role": SYSTEM_BUYER_ID, "obj": None},
         ]
         profiles.append(
             OneShotProfile(
@@ -814,16 +794,13 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         for i, (atype, aparams) in enumerate(zip(agent_types, agent_params)):
             a = instantiate(atype, **aparams)
             a.id = a.name
+            if a.adapted_object:
+                a.adapted_object.connect_to_adapter(a, None)
             self.join(a, i)
             agents.append(a)
-        self.agent_types = [
-            get_class(_["controller_type"])._type_name()
-            if _["controller_type"]
-            else "system_agent"
-            for _ in agent_params
-        ]
+        self.agent_types = [_.type_name for _ in agents]
         self.agent_params = [
-            {k: v for k, v in _.items() if k != "name" and k != "controller_type"}
+            {k: v for k, v in _.items() if k != "name" and k != "obj"}
             for _ in agent_params
         ]
         self.agent_unique_types = [
@@ -952,9 +929,7 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         self.exogenous_pin = defaultdict(int)
         self.exogenous_contracts_summary = None
 
-        self.initial_balances = dict(
-            zip(self.agents.keys(), initial_balance)
-        )
+        self.initial_balances = dict(zip(self.agents.keys(), initial_balance))
 
     @classmethod
     def generate(
@@ -971,11 +946,13 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         profit_means: Union[np.ndarray, Tuple[float, float], float] = (0.1, 0.2),
         profit_stddevs: Union[np.ndarray, Tuple[float, float], float] = 0.05,
         max_productivity: Union[np.ndarray, Tuple[float, float], float] = 1.0,
+        initial_balance: Optional[Union[np.ndarray, Tuple[int, int], int]] = None,
         cost_increases_with_level=True,
         equal_exogenous_supply=False,
         equal_exogenous_sales=False,
         exogenous_control: Union[Tuple[float, float], float] = -1,
-        force_signing=True,
+        cash_availability: Union[Tuple[float, float], float] = (1.5, 2.5),
+        force_signing=False,
         profit_basis=np.mean,
         storage_cost: Union[np.ndarray, Tuple[float, float], float] = (0.0, 0.2),
         delivery_penalty: Union[np.ndarray, Tuple[float, float], float] = (0.0, 0.2),
@@ -1017,6 +994,11 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             equal_exogenous_sales: If true, external sales will be distributed equally among all agents in the last
                                    layer
             force_signing: Whether to force contract signatures (exogenous contracts are treated in the same way).
+            exogenous_control: How much control does the agent have over exogenous contract signing. Only effective if
+                               force_signing is False and use_exogenous_contracts is True
+            cash_availability: The fraction of the total money needs of the agent to work at maximum capacity that
+                               is available as `initial_balance` . This is only effective if `initial_balance` is set
+                               to `None` .
             exogenous_control: How much control does the agent have over exogenous contract signing. Only effective if
                                force_signing is False and use_exogenous_contracts is True
             storage_cost: A range to sample mean-storage costs for all factories from
@@ -1053,12 +1035,13 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             profit_means=profit_means,
             profit_stddevs=profit_stddevs,
             max_productivity=max_productivity,
+            initial_balance=initial_balance,
             cost_increases_with_level=cost_increases_with_level,
             equal_exogenous_sales=equal_exogenous_sales,
             equal_exogenous_supply=equal_exogenous_supply,
+            cash_availability=cash_availability,
             price_multiplier=price_multiplier,
             exogenous_price_dev=exogenous_price_dev,
-            cash_availability=float("inf"),
             profit_basis="min"
             if profit_basis == np.min
             else "mean"
@@ -1093,6 +1076,8 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         assert n_agents >= n_processes
         n_products = n_processes + 1
         production_costs = make_array(production_costs, n_agents, dtype=int)
+        if initial_balance is not None:
+            initial_balance = make_array(initial_balance, n_agents, dtype=int)
         if not isinstance(agent_types, Iterable):
             agent_types = [agent_types] * n_agents
             if agent_params is None:
@@ -1126,7 +1111,7 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
 
         for t, p in zip(agent_types, agent_params):
             p["controller_type"] = t
-        agent_types = [_OneShotBaseAgent] * len(agent_types)
+        agent_types = [_OneShotAdapter] * len(agent_types)
         # generate production costs making sure that every agent can do exactly one process
         n_agents_cumsum = n_agents_per_process.cumsum().tolist()
         first_agent = [0] + n_agents_cumsum[:-1]
@@ -1324,6 +1309,17 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         )
 
         assert nxt == n_agents
+        if initial_balance is None:
+            # every agent at every level will have just enough to do all the needed to do cash_availability fraction of
+            # production (even though it may not have enough lines to do so)
+            cash_availability = realin(cash_availability)
+            balance = np.ceil(
+                np.sum(total_costs, axis=1) / n_agents_per_process
+            ).astype(int)
+            initial_balance = []
+            for b, a in zip(balance, n_agents_per_process):
+                initial_balance += [int(math.ceil(b * cash_availability))] * a
+        b = np.sum(initial_balance)
         info.update(
             dict(
                 product_prices=product_prices,
@@ -1337,8 +1333,12 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
                 expected_welfare=float(np.sum(max_income)),
                 expected_income_per_step=max_income.sum(axis=0),
                 expected_income_per_process=max_income.sum(axis=-1),
-                expected_mean_profit=float(np.sum(max_income)),
-                expected_profit_sum=float(n_agents * np.sum(max_income)),
+                expected_mean_profit=float(np.sum(max_income))
+                if b != 0
+                else np.sum(max_income),
+                expected_profit_sum=float(n_agents * np.sum(max_income) / b)
+                if b != 0
+                else n_agents * np.sum(max_income),
             )
         )
 
@@ -1440,15 +1440,18 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             exogenous_contracts=exogenous,
             agent_types=agent_types,
             agent_params=agent_params,
+            initial_balance=initial_balance,
             n_steps=n_steps,
             info=info,
-            price_multiplier=price_multiplier,
             force_signing=force_signing,
+            price_multiplier=price_multiplier,
+            inventory_valuation_trading=0,
+            inventory_valuation_catalog=0,
             **kwargs,
         )
 
     def add_financial_report(
-        self, agent: _OneShotBaseAgent, reports_agent, reports_time
+        self, agent: _OneShotAdapter, reports_agent, reports_time
     ) -> None:
         """
         Records a financial report for the given agent in the agent indexed reports and time indexed reports
@@ -1486,6 +1489,12 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             reports_time[str(self.current_step)] = {}
         reports_time[str(self.current_step)][agent.id] = report
 
+    def complete_contract_execution(self, *args, **kwargs):
+        pass
+
+    def start_contract_execution(self, contract: Contract) -> Optional[Set[Breach]]:
+        return set()
+
     def simulation_step(self, stage):
         s = self.current_step
         self.exogenous_qout = defaultdict(int)
@@ -1662,10 +1671,8 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             "id": contract.id,
             "seller_name": self.agents[contract.annotation["seller"]].name,
             "buyer_name": self.agents[contract.annotation["buyer"]].name,
-            "seller_type": self.agents[
-                contract.annotation["seller"]
-            ].__class__.__name__,
-            "buyer_type": self.agents[contract.annotation["buyer"]].__class__.__name__,
+            "seller_type": self.agents[contract.annotation["seller"]].type_name,
+            "buyer_type": self.agents[contract.annotation["buyer"]].type_name,
             "delivery_time": contract.agreement["time"],
             "quantity": contract.agreement["quantity"],
             "unit_price": contract.agreement["unit_price"],
@@ -1837,7 +1844,7 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         return [_ for _ in self.agents.keys() if is_system_agent(_)]
 
     @property
-    def non_system_agents(self) -> List[_OneShotBaseAgent]:
+    def non_system_agents(self) -> List[_OneShotAdapter]:
         """Returns all agents except system agents"""
         return [_ for _ in self.agents.values() if not is_system_agent(_.id)]
 
@@ -1932,7 +1939,7 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
             extra = dict()
         partners = [_ for _ in self.suppliers[product] if not self.is_bankrupt[_]]
         if not partners:
-            return False
+            return True
         if negotiators is None:
             negotiators = [
                 controller.create_negotiator(PassThroughSAONegotiator, name=_)
@@ -2037,7 +2044,7 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
         for aid, a in self.agents.items():
             if is_system_agent(aid):
                 continue
-            controllers[aid] = a.make_controller()
+            controllers[aid] = a.adapted_object
         for product in range(1, self.n_products):
             for aid in self.consumers[product]:
                 if is_system_agent(aid):
@@ -2076,3 +2083,8 @@ class SCML2020OneShotWorld(NoContractExecutionMixin, World):
 
     def get_private_state(self, agent: "Agent") -> dict:
         return agent.awi.state
+
+    def _contract_record(self, contract):
+        record = super()._contract_record(contract)
+        record["executed_at"] = self.current_step
+        return record
